package job

import (
	"context"
	"fmt"
	"sync"

	"github.com/google/uuid"
	"go.uber.org/zap"

	"github.com/freeman42332/zdb1000n/src/job/config"
	"github.com/freeman42332/zdb1000n/src/utils/metrics"
	"github.com/freeman42332/zdb1000n/src/utils/templates"
)

func sequenceJob(ctx context.Context, args config.Args, globalConfig *GlobalConfig, a *metrics.Accumulator, logger *zap.Logger) (data any, err error) {
	var jobConfig struct {
		BasicJobConfig

		Jobs []config.Config
	}

	if err := ParseConfig(&jobConfig, args, *globalConfig); err != nil {
		return nil, fmt.Errorf("error parsing job config: %w", err)
	}

	for _, cfg := range jobConfig.Jobs {
		job := Get(cfg.Type)
		if job == nil {
			return nil, fmt.Errorf("unknown job %q", cfg.Type)
		}

		data, err := job(ctx, cfg.Args, globalConfig, a, logger)
		if err != nil {
			return nil, fmt.Errorf("error running job: %w", err)
		}

		ctx = context.WithValue(ctx, templates.ContextKey("data."+cfg.Name), data)
	}

	return nil, nil
}

func parallelJob(ctx context.Context, args config.Args, globalConfig *GlobalConfig, a *metrics.Accumulator, logger *zap.Logger) (data any, err error) {
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()

	var jobConfig struct {
		BasicJobConfig

		Jobs []config.Config
	}

	if err := ParseConfig(&jobConfig, args, *globalConfig); err != nil {
		return nil, fmt.Errorf("error parsing job config: %w", err)
	}

	var wg sync.WaitGroup

	for i := range jobConfig.Jobs {
		job := Get(jobConfig.Jobs[i].Type)
		if job == nil {
			logger.Warning("Unknown job", zap.String("type", jobConfig.Jobs[i].Type))

			continue
		}

		if jobConfig.Jobs[i].Count < 1 {
			jobConfig.Jobs[i].Count = 1
		}

		for j := 0; j < jobConfig.Jobs[i].Count; j++ {
			wg.Add(1)

			go func(i int, a *metrics.Accumulator) {
				if _, err := job(ctx, jobConfig.Jobs[i].Args, globalConfig, a, logger); err != nil {
					logger.Error("error running one of the jobs", zap.Error(err))
				}

				wg.Done()
			}(i, a.Clone(uuid.NewString())) // metrics.Accumulator is not safe for concurrent use, so let's make a new one
		}
	}

	wg.Wait()

	return nil, nil
}
